
#define PIN_LOW1 7
#define PIN_LOW2 20
#define PIN_LOW3 49
#define PIN_LOW4 53



#define PIN_STEP1 10
#define PIN_DIR1 8

#define PIN_STEP2 35
#define PIN_DIR2 31

#define PIN_STEP3 18
#define PIN_DIR3 16

#define PIN_STEP4 45
#define PIN_DIR4 41

#define PIN_STEP5 27
#define PIN_DIR5 23

#define PIN_STEP6 12
#define PIN_DIR6 14

#define PIN_ESTOP_SENSOR 3









//****************************************************************************************************************
//    Port manipulation
//          
//****************************************************************************************************************


typedef void(*FP)(void);

FP step_high[7];
FP step_low[7];

FP dir_up[7];
FP dir_dn[7];






void step1_high() {
	PORTB |= (1 << PORTB4);
}
void step1_low() {
	PORTB &= ~(1 << PORTB4);
}

void step2_high() {
	PORTC |= (1 << PORTC2);
}
void step2_low() {
	PORTC &= ~(1 << PORTC2);
}


void step3_high() {
	PORTD |= (1 << PORTD3);
}
void step3_low() {
	PORTD &= ~(1 << PORTD3);
}

void step4_high() {
PORTL |= (1 << PORTL4);
}
void step4_low() {
	PORTL &= ~(1 << PORTL4);
}



void step5_high() {
	PORTA |= (1 << PORTA5);
}
void step5_low() {
	PORTA &= ~(1 << PORTA5);
}


void step6_high() {
	PORTB |= (1 << PORTB6);
}
void step6_low() {
	PORTB &= ~(1 << PORTB6);
}




void dir1_high() {
	PORTH |= (1 << PORTH5);
}
void dir1_low() {
	PORTH &= ~(1 << PORTH5);
}


void dir2_high() {
	PORTC |= (1 << PORTC6);
}
void dir2_low() {
	PORTC &= ~(1 << PORTC6);
}

void dir3_high() {
	PORTH |= (1 << PORTH1);
}
void dir3_low() {
	PORTH &= ~(1 << PORTH1);
}

void dir4_high() {
	PORTG |= (1 << PORTG0);
}
void dir4_low() {
	PORTG &= ~(1 << PORTG0);
}

void dir5_high() {
	PORTA |= (1 << PORTA1);
}
void dir5_low() {
	PORTA &= ~(1 << PORTA1);
}

void dir6_high() {
	PORTJ |= (1 << PORTJ1);
}
void dir6_low() {
	PORTJ &= ~(1 << PORTJ1);
}





void dir1_up() {
	dir1_low();
}

void dir2_up() {
	dir2_high();
}

void dir3_up() {
	dir3_high();
}

void dir4_up() {
	dir4_low();
}

void dir5_up() {
	dir5_high();
}

void dir6_up() {
	dir6_low();
}


void dir1_dn() {
	dir1_high();
}

void dir2_dn() {
	dir2_low();
}

void dir3_dn() {
	dir3_low();
}

void dir4_dn() {
	dir4_high();
}

void dir5_dn() {
	dir5_low();
}

void dir6_dn() {
	dir6_high();
}



//****************************************************************************************************************
//    Position control
//          
//****************************************************************************************************************


bool position_initialized = false;




int global_step_delay = 30000;

int threshold_position_offset = 600; // if position offset of any of the actuators exceeds this -- maximum of global_step_delay and smoothing_step_delay is used
int smoothing_step_delay = 500;

int max_position_offset = 0;


int pos[7];
int tgt[7];


//****************************************************************************************************************
//    Serial communications
//          
//****************************************************************************************************************

#define BAUDRATE 500000



#define START_BYTE '<'
#define END_BYTE '>'

// packet types: 
/// <5000,5000,5000,5000,5000,5000>	LEN 29 	--- set motor target positions

/// <S,0,0,0010>					LEN 10	--- control speed by setting inter-step delay to 10 microseconds
/// <M,1,U,0025>					LEN	10	--- move motor 1 up by 25 pulses
/// <N,0,0,5000>					LEN	10	--- set coordinate of the current position
/// <R,0,0,0000>          LEN 10  --- report motor positions
/// <E,0,0,0000>          LEN 10  --- report estop state

char packet[29];
int packet_cursor = 0;


void ParseCommandPacket() {
	char command_type = packet[0];
	char buf[5];

	int motor_id = 0;
	int dir = 0;

	int on_off = 0;

	int position_param = 0;

	String report = "";


	switch (command_type) {
	case 'S':
		// set step delay
		buf[0] = packet[6];
		buf[1] = packet[7];
		buf[2] = packet[8];
		buf[3] = packet[9];
		buf[4] = '\0';

		global_step_delay = atoi(buf) * 10;

		report = "Step delay set to " + String(global_step_delay);
		SendStatus(report);

		break;
	case 'M':
		// move individual motor incrementally
		motor_id = packet[2] - 48;

		dir = 1;
		if (packet[4] == 'D') {
			dir = -1;
		}

		buf[0] = packet[6];
		buf[1] = packet[7];
		buf[2] = packet[8];
		buf[3] = packet[9];
		buf[4] = '\0';

		position_param = atoi(buf);

		tgt[motor_id] += position_param * dir;

		report = "M" + String(motor_id) + " to move " + packet[4] + position_param;
		SendStatus(report);

		break;
	case 'N':
		// set coordinate of the current position
		motor_id = packet[2] - 48;

		buf[0] = packet[6];
		buf[1] = packet[7];
		buf[2] = packet[8];
		buf[3] = packet[9];
		buf[4] = '\0';

		position_param = atoi(buf);

		if (motor_id == 0) {

			for (int i = 1; i < 7; i++) {
				pos[i] = position_param;
				tgt[i] = position_param;
			}


		}
		else {

			pos[motor_id] = position_param;
			tgt[motor_id] = position_param;
		}


		position_initialized = true;

		report = "M" + String(motor_id) + " position set to " + position_param;
		SendStatus(report);

		break;

	case 'R':
		// report motor positions

		report = "Pos:" + String(pos[1]);

		for (int i = 2; i < 7; i++) {
			report += "," + String(pos[i]);
		}
		SendStatus(report);

	case 'E':
		// report estop state

		report = estop_report();

		SendStatus(report);
	}
}



void ParseTargetPositionPacket() {

	bool no_error = true;

	int tgt_buf[6];


	for (int i = 0; i < 6; i++) {
		char buf[5];
		buf[0] = packet[i * 5];
		buf[1] = packet[i * 5 + 1];
		buf[2] = packet[i * 5 + 2];
		buf[3] = packet[i * 5 + 3];
		buf[4] = '\0';

		tgt_buf[i] = atoi(buf);




		if (tgt_buf[i] < 0 || tgt_buf[i]>9999) {
			no_error = false;
			break;
		}
	}

	if (no_error) {
		for (int i = 0; i < 6; i++) {
			tgt[i + 1] = tgt_buf[i];
		}





	}

	packet_cursor = 0;

}




void CheckSerial() {

	if (Serial.available()) {
		char next_byte = Serial.read();

		if (next_byte == START_BYTE) {
			packet_cursor = 0;
		}
		else if (next_byte == END_BYTE) {
			if (packet_cursor == 29) {
				ParseTargetPositionPacket();
			}
			else if (packet_cursor == 10) {
				ParseCommandPacket();
			}
			else {
				packet_cursor = 0;
			}
		}
		else {
			packet[packet_cursor] = next_byte;
			packet_cursor++;
		}
	}

}


void SendStatus(String status_string) {

	Serial.print("<");
	Serial.print(status_string);
	Serial.print(">");

}


//****************************************************************************************************************
//    E-Stop
//          
//****************************************************************************************************************


bool estop = true;

String estop_report() {
	if (estop) {
		return "ESTOP";
	}
	else {
		return "NO ESTOP";
	}
}


#define ESTOP_READ_SAMPLES 10
int estop_pin_reads[ESTOP_READ_SAMPLES];
int estop_read_counter = 0;

#define ESTOP_STATE_SAMPLES 10
int estop_states[ESTOP_STATE_SAMPLES];
int estop_state_counter = 0;

void CheckEStop() {
	bool was_estop = estop;

	estop = true;

	int estop_sum = 0;

	estop_pin_reads[estop_read_counter] = bitRead(PINE, 5);
	estop_read_counter++;
	if (estop_read_counter == ESTOP_READ_SAMPLES) { estop_read_counter = 0; }


	for (int es = 0;es < ESTOP_READ_SAMPLES;es++) {
		estop_sum += estop_pin_reads[es];
	}

	if (estop_sum > 2) {
		estop_states[estop_state_counter] = 1;
		estop_state_counter++;
		if (estop_state_counter == ESTOP_STATE_SAMPLES) { estop_state_counter = 0; }

	}
	else {
		estop_states[estop_state_counter] = 0;
		estop_state_counter++;
		if (estop_state_counter == ESTOP_STATE_SAMPLES) { estop_state_counter = 0; }

	}

	estop_sum = 0;
	for (int es = 0;es < ESTOP_STATE_SAMPLES;es++) {
		estop_sum += estop_states[es];
	}

	if (estop_sum > 0) {
		estop = true;
		position_initialized = false;
		if (!was_estop) {
			SendStatus("ESTOP");
		}
	}
	else {
		estop = false;

		if (was_estop) {
			SendStatus("OK");
		}


	}


}


void setup() {


	pinMode(PIN_ESTOP_SENSOR, INPUT);
	digitalWrite(PIN_ESTOP_SENSOR, HIGH);



	pinMode(PIN_LOW1, OUTPUT);
	pinMode(PIN_LOW2, OUTPUT);
	pinMode(PIN_LOW3, OUTPUT);
	pinMode(PIN_LOW4, OUTPUT);

	step_high[1] = &step1_high;
	step_high[2] = &step2_high;
	step_high[3] = &step3_high;
	step_high[4] = &step4_high;
	step_high[5] = &step5_high;
	step_high[6] = &step6_high;

	step_low[1] = &step1_low;
	step_low[2] = &step2_low;
	step_low[3] = &step3_low;
	step_low[4] = &step4_low;
	step_low[5] = &step5_low;
	step_low[6] = &step6_low;

	dir_up[1] = &dir1_up;
	dir_up[2] = &dir2_up;
	dir_up[3] = &dir3_up;
	dir_up[4] = &dir4_up;
	dir_up[5] = &dir5_up;
	dir_up[6] = &dir6_up;

	dir_dn[1] = &dir1_dn;
	dir_dn[2] = &dir2_dn;
	dir_dn[3] = &dir3_dn;
	dir_dn[4] = &dir4_dn;
	dir_dn[5] = &dir5_dn;
	dir_dn[6] = &dir6_dn;




	DDRA = 0b11111111;
	DDRB = 0b11111111;
	DDRC = 0b11111111;
	DDRD = 0b11111111;
	DDRG = 0b11111111;
	DDRH = 0b11111111;
	DDRJ = 0b11111111;
	DDRL = 0b11111111;




	digitalWrite(PIN_LOW1, LOW);
	digitalWrite(PIN_LOW2, LOW);
	digitalWrite(PIN_LOW3, LOW);
	digitalWrite(PIN_LOW4, LOW);



	Serial.begin(BAUDRATE);
	Serial.setTimeout(10);


	for (int i = 1; i < 7; i++) {
		pos[i] = -1;
		tgt[i] = -1;
	}



}




unsigned long  prev_step_micros = 0;
void loop() {

	CheckEStop();

	CheckSerial();

	

	if ((!estop) && position_initialized) {

		if (prev_step_micros > micros()) {
			//happens when micros() overflow back to zero
			prev_step_micros = 0;
		}


		// check if position offset is too high and we need to limit movement speed to avoid sudden high amlitude jumps
		int local_step_delay = 30000;
		if (max_position_offset > threshold_position_offset) {
			// need to limit speed
			local_step_delay = max(global_step_delay, smoothing_step_delay);
		}
		else {
			// position offset is small, no speed limitation is needed
			local_step_delay = global_step_delay;
		}


		if ((micros() - prev_step_micros) > local_step_delay) {

			max_position_offset = 0;

			bool not_in_position = false;
			bool to_move[7];

			prev_step_micros = micros();


			not_in_position = false;
			for (int i = 1; i < 7; i++) {
				to_move[i] = false;
				if (tgt[i] != pos[i]) {
					not_in_position = true;
					to_move[i] = true;

					int position_offset = tgt[i] - pos[i];


					max_position_offset = max(abs(position_offset), max_position_offset);


					if (position_offset > 0) {
						dir_up[i]();
						pos[i]++;
					}
					else {
						dir_dn[i]();
						pos[i]--;
					}
				}
			}

			if (not_in_position) {

				for (int i = 1; i < 7; i++) {
					if (to_move[i]) {
						step_high[i]();
					}
				}

				delayMicroseconds(1);
			}

			for (int i = 1; i < 7; i++) {
				if (to_move[i]) {
					step_low[i]();
				}
			}

		}

	}



}

